import 'dart:typed_data';

import 'package:meta/meta.dart';
import 'package:bird_cryptography/bird_cryptography.dart';
import 'package:pointycastle/export.dart';

class PointycastleCryptographyKDF implements CryptographyKDF {
  const PointycastleCryptographyKDF();

  @override
  KDFCrypto scrypt({
    @required int n,
    @required int r,
    @required int p,
    @required int length,
  }) {
    assert(n != null);
    assert(r != null);
    assert(p != null);
    assert(length != null);
    return _AnonymousDerivator((List<int> salt, List<int> pw) => (Scrypt()
          ..init(ScryptParameters(n, r, p, length, Uint8List.fromList(salt))))
        .process(Uint8List.fromList(pw)));
  }

  @override
  KDFCrypto sha256_pbkdf2({
    @required int i,
    @required int length,
  }) {
    assert(i != null);
    assert(length != null);
    return _AnonymousDerivator((List<int> salt, List<int> pw) =>
        (PBKDF2KeyDerivator(HMac(SHA256Digest(), 64))
              ..init(Pbkdf2Parameters(
                Uint8List.fromList(salt),
                i,
                length,
              )))
            .process(Uint8List.fromList(pw)));
  }

  @override
  KDFCrypto sha384_pbkdf2({
    @required int i,
    @required int length,
  }) {
    assert(i != null);
    assert(length != null);
    return _AnonymousDerivator((List<int> salt, List<int> pw) =>
        (PBKDF2KeyDerivator(HMac(SHA384Digest(), 128))
              ..init(Pbkdf2Parameters(
                Uint8List.fromList(salt),
                i,
                length,
              )))
            .process(Uint8List.fromList(pw)));
  }

  @override
  KDFCrypto sha512_pbkdf2({
    @required int i,
    @required int length,
  }) {
    assert(i != null);
    assert(length != null);
    return _AnonymousDerivator((List<int> salt, List<int> pw) =>
        (PBKDF2KeyDerivator(HMac(SHA512Digest(), 128))
              ..init(Pbkdf2Parameters(
                Uint8List.fromList(salt),
                i,
                length,
              )))
            .process(Uint8List.fromList(pw)));
  }
}

class _AnonymousDerivator extends KDFCrypto {
  final List<int> Function(List<int> intList, List<int> pw) derive;

  const _AnonymousDerivator(this.derive);

  @override
  List<int> digestRaw(List<int> intList, List<int> pw) {
    return derive(intList, pw);
  }
}
