import 'dart:typed_data';

import 'package:bird/smart_json.dart';
import 'package:convert/convert.dart';
import 'package:bird_cryptography/bird_cryptography.dart';

/// Please take a look at the following resources before handling Ethereum keystores:
/// Broad overview here https://medium.com/@julien.maffre/what-is-an-ethereum-keystore-file-86c8c5917b97
/// More info here: https://github.com/hashcat/hashcat/issues/1228
/// Spec: https://github.com/ethereum/wiki/wiki/Web3-Secret-Storage-Definition
class EthereumKeystoreModel extends SmartJSON<EthereumKeystoreModel> {

  /// Currently supported version: 3.
  /// Here is the definition of version 3.
  /// https://github.com/ethereum/wiki/wiki/Web3-Secret-Storage-Definition
  final SmartInt version = SmartInt(null);

  /// [id]
  /// A UUID generated when the keystore is generated
  /// but a good convention is <uuid>.json, where <uuid> is the 128-bit UUID given to the
  /// secret key (a privacy-preserving proxy for the secret key's address).
  final SmartString id = SmartString(null);

  /// Instruction for the procedure with which the private key can be retrieved.
  ///
  /// NOTE: MEW creates keystore files with a capital C for crypto.
  final SmartTFrom<EthereumKeystoreCryptoModel> crypto = SmartTFrom(
      EthereumKeystoreCryptoModel());

  @override
  EthereumKeystoreModel creator() => EthereumKeystoreModel();

  @override
  JSONPrototypeCodec<EthereumKeystoreModel> tactic() =>
      JSONCanonicalMapTactic.ignoreOnMissingCanonical(
        {"version": version, "id": id, "crypto": crypto, "Crypto": crypto},
        {"version": version, "id": id, "crypto": crypto},
      );

  Uint8List decryptOrNull(Uint8List pw) {
    if (pw != null && pw.isNotEmpty) {
      return crypto.safeApply(() => null, (c) {
        final p = c.kdfparams;
        final derivedKey = p.deriveKey(pw);
        final text = c.ciphertext.nullVal();
        final params = c.cipherparams;

        if (c.isMacValid(derivedKey) && text != null) {
          return params.decipher(text, derivedKey);
        } else {
          return null;
        }
      });
    } else {
      return null;
    }
  }
}

/// Definition here: https://github.com/ethereum/wiki/wiki/Web3-Secret-Storage-Definition#alterations-from-version-2
// ignore: must_be_immutable
class EthereumKeystoreCryptoModel
    extends SmartJSON<EthereumKeystoreCryptoModel> {

  final SmartString ciphertext = SmartString(null);

  /// Supported ciphers:
  ///  "cipher": "aes-128-ctr",
  final SmartString cipher = SmartString(null);

  /// if [cipher] is "aes-128-ctr"
  /// cipherparams should contain:
  ///     "cipherparams": {
  ///         "iv": "0aea0ea70c06c8f5c1e0b39da6df9119"
  ///     },
  ///
  /// if [cipher] is "aes-128-cbc"
  ///    *currently not supported*
  SmartTFrom<SmartRedirect<EthereumCipherInterface>> _cipherparams;

  /// "scrypt" or "pbkdf2"
  /// The key for the cipher is the leftmost 16 bytes of the derived key, i.e. DK[0..15]
  final SmartString kdf = SmartString(null);

  /// if [kdf] is "scrypt":
  /// should contain
  ///      "dklen": 32,
  ///      "salt": "a0024a24fac530dd4493c407bf463430b6c6c9cf6f8879a6ad99309baa1f3163",
  ///      "n": 8192,
  ///      "r": 8,
  ///      "p": 1
  ///
  /// if [kdf] is "pbkdf2"
  /// should contain
  ///      "prf": "hmac-sha256" (may be extended in the future);
  ///      "c": number of iterations;
  ///      "salt": salt passed to PBKDF;
  ///      "dklen": length for the derived key. Must be >= 32.
  SmartTFrom<SmartRedirect<EthereumKDFInterface>> _kdfparams;

  /// "second leftmost 16 bytes" of the derived hash are concatenated with the ciphertext,
  /// which is then hashed with Keccak-256. The Keccak-256 hash is then directly compared
  /// to the "mac" value to verify whether or not the password is correct.
  final SmartString mac = SmartString(null);

  EthereumCipherInterface get cipherparams =>
      _cipherparams
          ?.nullVal()
          ?.access;

  EthereumKDFInterface get kdfparams =>
      _kdfparams
          ?.nullVal()
          ?.access;

  EthereumKeystoreCryptoModel() {
    _kdfparams = SmartTFrom(SmartRedirect<EthereumKDFInterface>((dynamic d) {
      if (kdf.nullVal() == "scrypt") {
        return EthereumKeystoreScryptKDFParams().from(d);
      } else if (kdf.nullVal() == "pbkdf2") {
        return EthereumKeystorePKDF2KDFParams().from(d);
      } else {
        return EthereumKeystoreUnknownKDFParams().from(d);
      }
    }));
    _cipherparams =
        SmartTFrom(SmartRedirect<EthereumCipherInterface>((dynamic d) {
          if (cipher.nullVal() == "aes-128-ctr") {
            return EthereumKeystoreAES128CTRParamsModel().from(d);
          } else {
            return EthereumKeystoreUnknownCipherParamsModel().from(d);
          }
        }));
  }

  @override
  EthereumKeystoreCryptoModel creator() => EthereumKeystoreCryptoModel();

  @override
  JSONPrototypeCodec<EthereumKeystoreCryptoModel> tactic() =>
      JSONMapTactic.ignoreMissing({
        "ciphertext": ciphertext,
        "cipher": cipher,
        "cipherparams": _cipherparams,
        "kdf": kdf,
        "kdfparams": _kdfparams,
        "mac": mac,
      });

  bool isMacValid(List<int> derivedKey) {
    if (derivedKey == null) {
      return false;
    } else {
      return mac.nullVal() == deriveMac(derivedKey).printToTypeString(e_hex);
    }
  }

  FocusedCodec deriveMac(List<int> derivedKey) {
    if (derivedKey == null) {
      return null;
    } else {
      return FocusedCodec.raw(
          e_bytelist, CryptographyHashes.dart.keccak256().digestRaw(e_bytelist
          .decodedData(macBody(derivedKey))
          .list));
    }
  }

  List<int> macBody(List<int> derivedKey) {
    if (derivedKey == null) {
      return null;
    } else {
      return [derivedKey.sublist(16, 32), e_hex.decode(ciphertext.nullVal())]
          .expand((a) => a)
          .toList();
    }
  }
}

/// KDF
abstract class EthereumKDFInterface<T extends SmartJSON<T>>
    extends SmartJSON<T> {

  /// Check whether derivation with the current data is supported or not
  /// false means either the kdf function is unknown or params are invalid
  bool validParameters();

  /// Derives a key using the defined kdf.
  Uint8List deriveKey(List<int> pw);
}

class EthereumKeystoreUnknownKDFParams
    extends EthereumKDFInterface<EthereumKeystoreUnknownKDFParams> {
  @override
  EthereumKeystoreUnknownKDFParams creator() =>
      EthereumKeystoreUnknownKDFParams();

  @override
  JSONPrototypeCodec<EthereumKeystoreUnknownKDFParams> tactic() =>
      JSONMapTactic.ignoreMissing({});

  @override
  bool validParameters() => false;

  @override
  Uint8List deriveKey(List<int> pw) => null;
}

class EthereumKeystoreScryptKDFParams
    extends EthereumKDFInterface<EthereumKeystoreScryptKDFParams> {
  final SmartInt dklen = SmartInt(null);
  final SmartString salt = SmartString(null);
  final SmartInt n = SmartInt(null);
  final SmartInt r = SmartInt(null);
  final SmartInt p = SmartInt(null);

  @override
  EthereumKeystoreScryptKDFParams creator() =>
      EthereumKeystoreScryptKDFParams();

  @override
  JSONPrototypeCodec<EthereumKeystoreScryptKDFParams> tactic() =>
      JSONMapTactic.ignoreMissing(
          {"dklen": dklen, "salt": salt, "n": n, "r": r, "p": p});

  @override
  bool validParameters() =>
      dklen.nullVal() != null && salt.nullVal() != null && n.nullVal() != null
          && r.nullVal() != null && p.nullVal() != null;

  @override
  Uint8List deriveKey(List<int> pw) {
    if (pw != null && validParameters()) {
      return Uint8List.fromList(CryptographyKDF.dart.scrypt(
        length: dklen.nullVal(),
        n: n.nullVal(),
        p: p.nullVal(),
        r: r.nullVal(),
      )
          .digestRaw(Uint8List.fromList(hex.decode(salt.nullVal())), pw));
    } else {
      return null;
    }
  }
}

class EthereumKeystorePKDF2KDFParams
    extends EthereumKDFInterface<EthereumKeystorePKDF2KDFParams> {
  final SmartString prf = SmartString(null);
  final SmartInt c = SmartInt(null);
  final SmartString salt = SmartString(null);
  final SmartInt dklen = SmartInt(null);

  @override
  EthereumKeystorePKDF2KDFParams creator() => EthereumKeystorePKDF2KDFParams();

  @override
  JSONPrototypeCodec<EthereumKeystorePKDF2KDFParams> tactic() =>
      JSONMapTactic.ignoreMissing(
          {"prf": prf, "c": c, "salt": salt, "dklen": dklen});

  @override
  bool validParameters() =>
      prf.nullVal() != null && c.nullVal() != null && salt.nullVal() != null
          && dklen.nullVal() != null;

  @override
  Uint8List deriveKey(List<int> pw) {
    if (pw != null && validParameters()) {
      if (prf.nullVal() == "hmac-sha256") {
        return Uint8List.fromList(CryptographyKDF.dart.sha256_pbkdf2(
          i: c.nullVal(),
          length: dklen.nullVal(),
        ).digestRaw(Uint8List.fromList(hex.decode(salt.nullVal())), pw));
      } else {
        return null;
      }
    } else {
      return null;
    }
  }
}

/// CIPHER
abstract class EthereumCipherInterface
    extends SmartJSON<EthereumCipherInterface> {
  /// Check whether derivation with the current data is supported or not
  /// false means either the cipher function is unknown or params are invalid
  bool validParameters();

  /// Deciphers the [ciphertext] with the [derivedKey].
  Uint8List decipher(String ciphertext, Uint8List derivedKey);
}

class EthereumKeystoreUnknownCipherParamsModel extends EthereumCipherInterface {

  @override
  EthereumKeystoreUnknownCipherParamsModel creator() =>
      EthereumKeystoreUnknownCipherParamsModel();

  @override
  JSONPrototypeCodec<EthereumCipherInterface> tactic() =>
      JSONMapTactic.ignoreMissing({});

  @override
  Uint8List decipher(String ciphertext, Uint8List derivedKey) => null;

  @override
  bool validParameters() => false;
}

class EthereumKeystoreAES128CTRParamsModel extends EthereumCipherInterface {
  final SmartString iv = SmartString(null);

  @override
  EthereumKeystoreAES128CTRParamsModel creator() =>
      EthereumKeystoreAES128CTRParamsModel();

  @override
  JSONPrototypeCodec<EthereumCipherInterface> tactic() =>
      JSONMapTactic.ignoreMissing({"iv": iv});

  @override
  Uint8List decipher(String ciphertext, Uint8List derivedKey) {
    if (ciphertext != null && derivedKey != null && validParameters()) {
      return Uint8List.fromList(CryptographyCiphers.dart.aes_ctr().decrypt(
        FocusedCodec
            .raw(const CryptoEncodingIntList(), e_hex.decode(ciphertext))
            .list,
        CryptoParametersPWAndIV.create(
          FocusedCodec
              .raw(const CryptoEncodingIntList(), derivedKey.sublist(0, 16))
              .list,
          FocusedCodec
              .encoded(const CryptoEncodingHex(), iv.nullVal())
              .list,
        ),
      ))
//                .rawListOfInts
//                .fold(() => null, (a) => Uint8List.fromList(a))
          ;
    } else {
      return null;
    }
  }

  @override
  bool validParameters() {
    return iv.nullVal() != null && e_hex
        .decode(iv.nullVal())
        ?.length == 16;
  }
}