import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';

import 'package:meta/meta.dart';
import 'package:bird_cryptography/bird_cryptography.dart';
import 'package:test/test.dart';

const String keystorePath = "test/test_resources/cryptography/ethereum/";

/// https://github.com/ethereum/wiki/wiki/Web3-Secret-Storage-Definition
void main() {
    group("$EthereumKeystoreModel", () {
        group("1: Official TV 1", () {
            final _KeystoreTestVector officialTV1 = _KeystoreTestVector(
                password: "testpassword",
                derivedKey: "f06d69cdc7da0faffb1008270bca38f5e31891a3a773950e6d0fea48a7188551",
                macBody: "e31891a3a773950e6d0fea48a71885515318b4d5bcd28de64ee5559e671353e16f075ecae9f99c7a79a38af5f869aa46",
                cipherKey: "f06d69cdc7da0faffb1008270bca38f5",
                secret: "7a28b5ba57c53603b0b07b56bba752f7784bf506fa95edc395f5cf6c7514fe9d",
                keystoreJSON: json.decode(File(keystorePath + 'keystore_tv_1.json').readAsStringSync()),
            );
            officialTV1.testDerivations();
        });

        group("1: invalid cipher", () {
            final _KeystoreTestVector officialTV1invalid = _KeystoreTestVector(
                password: "testpassword",
                derivedKey: "f06d69cdc7da0faffb1008270bca38f5e31891a3a773950e6d0fea48a7188551",
                macBody: "e31891a3a773950e6d0fea48a71885515318b4d5bcd28de64ee5559e671353e16f075ecae9f99c7a79a38af5f869aa46",
                cipherKey: "f06d69cdc7da0faffb1008270bca38f5",
                secret: null,
                keystoreJSON: json.decode(File(keystorePath + 'keystore_tv_1_invalid.json').readAsStringSync()),
            );
            officialTV1invalid.testDerivations();
        });

        /// takes too long
//        group("2: Official TV 2", () {
//            final _KeystoreTestVector officialTV2 = _KeystoreTestVector(
//                password: "testpassword",
//                derivedKey: "fac192ceb5fd772906bea3e118a69e8bbb5cc24229e20d8766fd298291bba6bd",
//                macBody: "bb5cc24229e20d8766fd298291bba6bdd172bf743a674da9cdad04534d56926ef8358534d458fffccd4e6ad2fbde479c",
//                cipherKey: "fac192ceb5fd772906bea3e118a69e8b",
//                secret: "7a28b5ba57c53603b0b07b56bba752f7784bf506fa95edc395f5cf6c7514fe9d",
//                keystoreJSON: json.decode(File(keystorePath + 'keystore_tv_2.json').readAsStringSync()),
//            );
//            officialTV2.testDerivations(shouldBeInvalid: false);
//        });

        group("3: Scrypt & aes-128-ctr", () {
            final _KeystoreTestVector tv3 = _KeystoreTestVector(
                password: "testtesttest",
                derivedKey: "1e85c76114322482faa77459a67506d8c17c403cf003c284891406fe34ae46cd",
                macBody: "c17c403cf003c284891406fe34ae46cd" "670786ed00b108f2b74bdc0edaac28aec5d291919c270cf4d768da42e706db5b",
                cipherKey: "1e85c76114322482faa77459a67506d8",
                secret: "678c3b751515990f27eccb95d778414b1f6d9aa603ce060d28d2b4483bcb030f",
                keystoreJSON: json.decode(File(keystorePath + 'keystore_tv_3.json').readAsStringSync()),
            );
            tv3.testDerivations();
        });

        group("4: Scrypt & aes-128-ctr (3 with big C in Crypto)", () {
            final _KeystoreTestVector tv4 = _KeystoreTestVector(
                password: "testtesttest",
                derivedKey: "1e85c76114322482faa77459a67506d8c17c403cf003c284891406fe34ae46cd",
                macBody: "c17c403cf003c284891406fe34ae46cd" "670786ed00b108f2b74bdc0edaac28aec5d291919c270cf4d768da42e706db5b",
                cipherKey: "1e85c76114322482faa77459a67506d8",
                secret: "678c3b751515990f27eccb95d778414b1f6d9aa603ce060d28d2b4483bcb030f",
                keystoreJSON: json.decode(File(keystorePath + 'keystore_tv_4.json').readAsStringSync()),
            );
            tv4.testDerivations();
        });

        group("5: Scrypt, aes-128-ctr", () {
            final _KeystoreTestVector tv5 = _KeystoreTestVector(
                password: '''!"§\$%&/()''',
                derivedKey: "c459c8fa194868f2bb9b66fbf690ff1883e099b5a18e088514d4d5d0a9eb8480",
                macBody: "83e099b5a18e088514d4d5d0a9eb8480" "93f0a8cea8c65840f39c016e9862ab3713ae14458b5aee2e9161adc0535d9440",
                cipherKey: "c459c8fa194868f2bb9b66fbf690ff18",
                secret: "5910af3046baf6c80fdc7bfa56155ba506e42f68ee848a26c9b0b3ed5a0fa77b",
                keystoreJSON: json.decode(File(keystorePath + 'keystore_tv_5.json').readAsStringSync()),
            );
            tv5.testDerivations();
        });

        group("5: invalid kdf", () {
            /// Invalid kdf
            final _KeystoreTestVector tv5invalid = _KeystoreTestVector(
                password: '''!"§\$%&/()''',
                derivedKey: null,
                macBody: null,
                cipherKey: null,
                secret: null,
                isMacValid: false,
                keystoreJSON: json.decode(File(keystorePath + 'keystore_tv_5_invalid.json').readAsStringSync()),
            );
            tv5invalid.testDerivations();
        });
    });

    group("$EthereumKeystoreCryptoModel", () {
        /// Covered by keystore tests.
    });

    group("$EthereumCipherInterface", () {
        /// Nothing to test.
    });

    group("$EthereumKDFInterface", () {
        /// Nothing to test.
    });

    test("$EthereumKeystoreUnknownKDFParams", () {
        final EthereumKeystoreUnknownKDFParams params = EthereumKeystoreUnknownKDFParams();

        expect(params.validParameters(), false);
        expect(params.deriveKey([1]), null);
    });

    group("$EthereumKeystoreScryptKDFParams", () {
        /// Covered by keystore tests.
    });

    group("$EthereumKeystorePKDF2KDFParams", () {
        /// Covered by keystore tests.
    });

    group("$EthereumCipherInterface", () {
        /// Nothing to test.
    });

    test("$EthereumKeystoreUnknownCipherParamsModel", () {
        final EthereumKeystoreUnknownCipherParamsModel params = EthereumKeystoreUnknownCipherParamsModel();

        expect(params.validParameters(), false);
        expect(params.decipher("-", Uint8List.fromList([1])), null);
    });

    group("$EthereumKeystoreAES128CTRParamsModel", () {
        /// Covered by keystore tests.
    });
}

class _KeystoreTestVector {
    final dynamic keystoreJSON;
    final EthereumKeystoreModel keystore;
    final String password;
    final bool isMacValid;
    final String derivedKey;
    final String macBody;
    final String cipherKey;
    final String secret;

    _KeystoreTestVector({
        @required this.keystoreJSON,
        @required this.password,
        @required this.derivedKey,
        @required this.macBody,
        @required this.cipherKey,
        @required this.secret,
        this.isMacValid = true,
    }) : this.keystore = EthereumKeystoreModel().from(keystoreJSON);

    void testDerivations() {
        final pw = e_utf8.decode(password);
        final _derivedKey = keystore.crypto
            .nullVal()
            .kdfparams
            .deriveKey(pw);

        test("derivedKey", () {
            expect(e_hex.encode(_derivedKey), derivedKey);
        });

        test("mac valid", () {
            expect(keystore.crypto.nullVal().isMacValid(_derivedKey), isMacValid);
        });

        test("macBody", () {
            expect(e_hex.encode(keystore.crypto.nullVal().macBody(_derivedKey)), macBody);
        });

        test("cipher", () {
            expect(e_hex.encode(_derivedKey?.sublist(0, 16)), cipherKey);
        });

        test("secret", () {
            expect(e_hex.encode(keystore.decryptOrNull(Uint8List.fromList(pw))), secret);
        });

        test("decrypting wrong pw", () {
            expect(keystore.decryptOrNull(Uint8List.fromList(utf8.encode("wrongPassword"))), null);
        });

        test("decrypting empty pw", () {
            expect(keystore.decryptOrNull(Uint8List.fromList(utf8.encode(""))), null);
            expect(keystore.decryptOrNull(null), null);
        });
    }
}